@@ -661,10 +661,10 @@ garcon_gtk_menu_reload (GarconGtkMenu *menu)
 
 
 
-static GtkWidget*
-garcon_gtk_menu_load_icon (const gchar *icon_name)
+static GdkPixbuf*
+garcon_gtk_menu_load_icon_pixbuf (const gchar *icon_name,
+                                  gint         scale_factor)
 {
-  GtkWidget *image = NULL;
   gint w, h, size;
   gchar *p, *name = NULL;
   GdkPixbuf *pixbuf = NULL;
@@ -675,13 +675,15 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
 
   if (gtk_icon_theme_has_icon (icon_theme, icon_name))
     {
-	  pixbuf = gtk_icon_theme_load_icon (icon_theme, icon_name, size, 0, NULL);;
+      pixbuf = gtk_icon_theme_load_icon (icon_theme,
+                                         icon_name, size * scale_factor,
+                                         GTK_ICON_LOOKUP_FORCE_SIZE, NULL);
     }
   else
     {
       if (g_path_is_absolute (icon_name))
         {
-          pixbuf = gdk_pixbuf_new_from_file_at_scale (icon_name, w, h, TRUE, NULL);
+          pixbuf = gdk_pixbuf_new_from_file_at_scale (icon_name, w * scale_factor, h * scale_factor, TRUE, NULL);
         }
       else
         {
@@ -690,12 +692,14 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
           if (p)
             {
               name = g_strndup (icon_name, p - icon_name);
-              pixbuf = gtk_icon_theme_load_icon (icon_theme, name, size, 0, NULL);
+              pixbuf = gtk_icon_theme_load_icon (icon_theme,
+                                                 icon_name, size * scale_factor,
+                                                 GTK_ICON_LOOKUP_FORCE_SIZE, NULL);
               g_free (name);
               name = NULL;
             }
 
-          /* maybe they point to a file in the pixbufs folder */
+          /* maybe they point to a file in the pixmaps folder */
           if (G_UNLIKELY (pixbuf == NULL))
             {
               gchar *filename;
@@ -707,21 +711,32 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
 
           if (name)
             {
-              pixbuf = gdk_pixbuf_new_from_file_at_scale (name, w, h, TRUE, NULL);
+              pixbuf = gdk_pixbuf_new_from_file_at_scale (name, w * scale_factor, h * scale_factor, TRUE, NULL);
               g_free (name);
             }
         }
     }
 
-  /* Turn the pixbuf into a gtk_image */
-  if (G_LIKELY (pixbuf))
+  return pixbuf;
+}
+
+
+
+static GtkWidget*
+garcon_gtk_menu_load_icon (const gchar *icon_name,
+                           gint         scale_factor)
+{
+  GtkWidget *image = NULL;
+  GdkPixbuf *pixbuf = NULL;
+
+  pixbuf = garcon_gtk_menu_load_icon_pixbuf (icon_name, scale_factor);
+  if (G_LIKELY (pixbuf != NULL))
     {
-      /* scale the pixbuf down if it needs it */
-      GdkPixbuf *pixbuf_scaled = gdk_pixbuf_scale_simple (pixbuf, w, h, GDK_INTERP_BILINEAR);
+      cairo_surface_t *surface = gdk_cairo_surface_create_from_pixbuf (pixbuf, scale_factor, NULL);
       g_object_unref (G_OBJECT (pixbuf));
 
-      image = gtk_image_new_from_pixbuf (pixbuf_scaled);
-      g_object_unref (G_OBJECT (pixbuf_scaled));
+      image = gtk_image_new_from_surface (surface);
+      cairo_surface_destroy (surface);
     }
   else
     {
@@ -734,17 +749,67 @@ garcon_gtk_menu_load_icon (const gchar *icon_name)
 
 
 
+static void
+garcon_gtk_menu_item_scale_factor_changed (GtkWidget   *menu_item,
+                                           GParamSpec  *pspec,
+                                           const gchar *icon_name)
+{
+  GdkPixbuf *pixbuf;
+  gint       scale_factor;
+
+  g_return_if_fail (GTK_IS_MENU_ITEM (menu_item));
+  g_return_if_fail (icon_name != NULL);
+
+  scale_factor = gtk_widget_get_scale_factor (menu_item);
+  pixbuf = garcon_gtk_menu_load_icon_pixbuf (icon_name, scale_factor);
+
+  if (G_LIKELY (pixbuf))
+    {
+      GtkWidget *image = NULL;
+      GtkWidget *box;
+
+      box = gtk_bin_get_child (GTK_BIN (menu_item));
+      if (GTK_IS_BOX (box))
+        {
+          GList *children = gtk_container_get_children (GTK_CONTAINER (box));
+          for (GList *l = children; l != NULL; l = l->next)
+            {
+              if (GTK_IS_IMAGE (l->data))
+                {
+                  image = GTK_WIDGET (l->data);
+                  break;
+                }
+            }
+          g_list_free (children);
+        }
+
+      if (image != NULL)
+        {
+          GdkWindow       *window = gtk_widget_get_window (menu_item);
+          cairo_surface_t *surface = gdk_cairo_surface_create_from_pixbuf (pixbuf, scale_factor, window);
+          gtk_image_set_from_surface (GTK_IMAGE (image), surface);
+          cairo_surface_destroy (surface);
+        }
+
+      g_object_unref (G_OBJECT (pixbuf));
+    }
+}
+
+
+
 static GtkWidget*
 garcon_gtk_menu_create_menu_item (gboolean     show_menu_icons,
                                   const gchar *name,
-                                  const gchar *icon_name)
+                                  const gchar *icon_name,
+                                  gint         scale_factor)
 {
   GtkWidget *mi;
   GtkWidget *image;
+  gchar *icon_name_copy;
 
   if (show_menu_icons)
     {
-      image = garcon_gtk_menu_load_icon (icon_name);
+      image = garcon_gtk_menu_load_icon (icon_name, scale_factor);
       gtk_widget_show (image);
     }
   else
@@ -754,6 +819,11 @@ garcon_gtk_menu_create_menu_item (gboolean     show_menu_icons,
 
   mi = xfce_gtk_image_menu_item_new (name, NULL, NULL, NULL, NULL, image, NULL);
 
+  icon_name_copy = g_strdup (icon_name);
+  g_object_set_data_full (G_OBJECT (mi), "--garcon-icon-name", icon_name_copy, g_free);
+  g_signal_connect (mi, "notify::scale-factor",
+                    G_CALLBACK (garcon_gtk_menu_item_scale_factor_changed), icon_name_copy);
+
   return mi;
 }
 
@@ -764,7 +834,8 @@ garcon_gtk_menu_pack_actions_menu (GtkWidget      *menu,
                                    GarconMenuItem *menu_item,
                                    GList          *actions,
                                    const gchar    *parent_icon_name,
-                                   gboolean        show_menu_icons)
+                                   gboolean        show_menu_icons,
+                                   gint            scale_factor)
 {
   GList     *iter;
   GtkWidget *mi;
@@ -791,7 +862,8 @@ garcon_gtk_menu_pack_actions_menu (GtkWidget      *menu,
 
       mi = garcon_gtk_menu_create_menu_item (show_menu_icons,
                                              garcon_menu_item_action_get_name (action),
-                                             action_icon_name);
+                                             action_icon_name,
+                                             scale_factor);
 
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), mi);
       g_signal_connect_object (G_OBJECT (mi), "activate",
@@ -812,13 +884,14 @@ garcon_gtk_menu_add_actions (GarconGtkMenu  *menu,
                              const gchar    *parent_icon_name)
 {
   GtkWidget *submenu, *mi;
+  gint scale_factor = gtk_widget_get_scale_factor (GTK_WIDGET (menu));
 
   submenu = gtk_menu_new ();
 
   /* Add the parent item again, this time something the user can click to execute */
   mi = garcon_gtk_menu_create_menu_item (menu->priv->show_menu_icons,
                                          garcon_menu_item_get_name (menu_item),
-                                         parent_icon_name);
+                                         parent_icon_name, scale_factor);
   gtk_menu_shell_append (GTK_MENU_SHELL (submenu), mi);
 
   /* we need to store the GarconGtkMenu with this item so we can
@@ -829,7 +902,8 @@ garcon_gtk_menu_add_actions (GarconGtkMenu  *menu,
   gtk_widget_show (mi);
 
   garcon_gtk_menu_pack_actions_menu (submenu, menu_item, actions,
-                                     parent_icon_name, menu->priv->show_menu_icons);
+                                     parent_icon_name, menu->priv->show_menu_icons,
+                                     scale_factor);
 
   return submenu;
 }
@@ -920,6 +994,7 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
 {
   GList               *elements, *li;
   GtkWidget           *mi;
+  gint                 scale_factor;
   const gchar         *name, *icon_name;
   const gchar         *comment;
   GtkWidget           *submenu;
@@ -930,6 +1005,8 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
   g_return_if_fail (GTK_IS_MENU (gtk_menu));
   g_return_if_fail (GARCON_IS_MENU (garcon_menu));
 
+  scale_factor = gtk_widget_get_scale_factor (GTK_WIDGET (menu));
+
   elements = garcon_menu_get_elements (garcon_menu);
   for (li = elements; li != NULL; li = li->next)
     {
@@ -962,7 +1039,7 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
             icon_name = "applications-other";
 
           /* build the menu item */
-          mi = garcon_gtk_menu_create_menu_item (menu->priv->show_menu_icons, name, icon_name);
+          mi = garcon_gtk_menu_create_menu_item (menu->priv->show_menu_icons, name, icon_name, scale_factor);
           gtk_menu_shell_append (GTK_MENU_SHELL (gtk_menu), mi);
 
           /* if the menu item has actions such as "Private browsing mode"
@@ -1047,7 +1124,7 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
                 icon_name = "applications-other";
 
               /* build the menu item */
-              mi = garcon_gtk_menu_create_menu_item (menu->priv->show_menu_icons, name, icon_name);
+              mi = garcon_gtk_menu_create_menu_item (menu->priv->show_menu_icons, name, icon_name, scale_factor);
 
               gtk_menu_shell_append (GTK_MENU_SHELL (gtk_menu), mi);
               gtk_menu_item_set_submenu (GTK_MENU_ITEM (mi), submenu);
@@ -1065,8 +1142,8 @@ garcon_gtk_menu_add (GarconGtkMenu *menu,
 
 /**
  * garcon_gtk_menu_new:
- * @garcon_menu (nullable): The #GarconMenu to be associated with the
- *                          #GarconGtkMenu, or %NULL.
+ * @garcon_menu: (nullable): The #GarconMenu to be associated with the
+ *                           #GarconGtkMenu, or %NULL.
  *
  * Creates a new #GarconGtkMenu for the .menu file referred to by @garcon_menu.
  * This operation only fails if @garcon_menu is invalid.
@@ -1225,6 +1302,7 @@ garcon_gtk_menu_get_show_menu_icons (GarconGtkMenu *menu)
 /**
  * garcon_gtk_menu_set_show_tooltips:
  * @menu  : A #GarconGtkMenu
+ * @show_tooltips : new value
  *
  *
  **/
@@ -1314,13 +1392,15 @@ garcon_gtk_menu_get_desktop_actions_menu (GarconMenuItem *item)
   GList       *actions = NULL;
   const gchar *parent_icon_name;
   gboolean     show_menu_icons = FALSE;
+  gint         scale_factor;
 
   actions = garcon_menu_item_get_actions (item);
   g_return_val_if_fail (actions != NULL, NULL);
 
   parent_icon_name = garcon_menu_item_get_icon_name (item);
+  scale_factor = gtk_widget_get_scale_factor (GTK_WIDGET (submenu));
 
-  garcon_gtk_menu_pack_actions_menu (submenu, item, actions, parent_icon_name, show_menu_icons);
+  garcon_gtk_menu_pack_actions_menu (submenu, item, actions, parent_icon_name, show_menu_icons, scale_factor);
   g_list_free (actions);
 
   return GTK_MENU (submenu);
